# -*- coding: utf-8 -*-

# -------------------------------------------------------------------------------
# Author:       CAO.T.F
# Date:         2021-04-19 11:37
# Description:  重写logging.FileHandler，支持多进程的TimedRotatingFileHandler

# -------------------------------------------------------------------------------
import os
import re
import datetime
import logging


class MultiprocessHandler(logging.FileHandler):
    def __init__(self, filename, when='D', backupCount=0, encoding=None, delay=False):
        """
        :param filename: 日志文件名
        :param when: 日志文件切分时间间隔的单位
        :param backupCount: 保留文件个数
        :param encoding: 文件编码格式
        :param delay: 是否开启 OutSteam缓存。
                        True 表示开启缓存，OutStream 输出到缓存，待缓存区满后，刷新缓存区，并输出缓存数据到文件。
                        False表示不缓存，OutStream 直接输出到文件
        """
        self.prefix = filename
        self.backupCount = backupCount
        self.when = when.upper()
        # 正则匹配 年-月-日
        self.extMath = r"^\d{4}-\d{2}-\d{2}"

        self.when_dict = {
            'S': "%Y-%m-%d-%H-%M-%S",  # S 每秒建立一个新文件
            'M': "%Y-%m-%d-%H-%M",  # M 每分钟建立一个新文件
            'H': "%Y-%m-%d-%H",  # H 每天建立一个新文件
            'D': "%Y-%m-%d"  # D 每天建立一个新文件
        }
        # 日志文件日期后缀
        self.suffix = self.when_dict.get(when)
        if not self.suffix:
            raise ValueError(u"指定的日期间隔单位无效: %s" % self.when)

        # 拼接文件路径，格式化字符串
        self.file_formatter = os.path.join("logs", "%s.%s" % (self.prefix, self.suffix))
        # 使用当前时间，格式化文件格式化字符串
        self.file_path = datetime.datetime.now().strftime(self.file_formatter)
        # 获得文件夹路径
        _dir = os.path.dirname(self.file_formatter)
        try:
            # 如果日志文件夹不存在，则创建文件夹
            if not os.path.exists(_dir):
                os.makedirs(_dir)
        except Exception as mk_dir_exception:
            print(mk_dir_exception)

        logging.FileHandler.__init__(self, self.file_path, 'a+', encoding, delay)

    def write_log_to_file(self):
        """
        更改日志写入目的写入文件
        :return True 表示已更改，False 表示未更改
        """
        # 以当前时间获得新日志文件路径
        _file_path = datetime.datetime.now().strftime(self.file_formatter)
        # 新日志文件日期 不等于 旧日志文件日期，则表示 已经到了日志切分的时候
        # 更换日志写入目的为新日志文件。
        # 例如 按 天 （D）来切分日志
        #   当前新日志日期等于旧日志日期，则表示在同一天内，还不到日志切分的时候
        #   当前新日志日期不等于旧日志日期，则表示不在
        # 同一天内，进行日志切分，将日志内容写入新日志内。
        if _file_path != self.file_path:
            self.file_path = _file_path
            return True
        return False

    def do_log_file_backup(self):
        """
        输出信息到日志文件，并删除多于保留个数的所有日志文件
        """
        # 日志文件的绝对路径
        self.baseFilename = os.path.abspath(self.file_path)
        # stream == OutStream
        # stream is not None 表示 OutStream中还有未输出完的缓存数据
        if self.stream:
            # flush close 都会刷新缓冲区，flush不会关闭stream，close则关闭stream
            # self.stream.flush()
            self.stream.close()
            # 关闭stream后必须重新设置stream为None，否则会造成对已关闭文件进行IO操作。
            self.stream = None
        # delay 为False 表示 不OutStream不缓存数据 直接输出
        #   所有，只需要关闭OutStream即可
        if not self.delay:
            # 这个地方如果关闭colse那么就会造成进程往已关闭的文件中写数据，从而造成IO错误
            # delay == False 表示的就是 不缓存直接写入磁盘
            # 我们需要重新在打开一次stream
            # self.stream.close()
            self.stream = self._open()
        # 删除多于保留个数的所有日志文件
        if self.backupCount > 0:
            for s in self.get_file_to_remove():
                os.remove(s)

    def get_file_to_remove(self):
        """
        获得过期需要删除的日志文件
        """
        # 分离出日志文件夹绝对路径
        # _ 表示占位符，没什么实际意义，
        dir_name, _ = os.path.split(self.baseFilename)
        file_names = os.listdir(dir_name)
        result = []
        # self.prefix 为日志文件名 列如：log.2017-03-19 中的 log
        # 加上 点号 . 方便获取点号后面的日期
        prefix = self.prefix + '.'
        prefix_len = len(prefix)
        for file_name in file_names:
            if file_name[:prefix_len] == prefix:
                # 日期后缀 log.2017-03-19 中的 2017-03-19
                suffix = file_name[prefix_len:]
                # 匹配符合规则的日志文件，添加到result列表中
                if re.compile(self.extMath).match(suffix):
                    result.append(os.path.join(dir_name, file_name))
        result.sort()

        # 返回 待删除的日志文件
        #   多于 保留文件个数 backupCount的所有前面的日志文件。
        if len(result) < self.backupCount:
            result = []
        else:
            result = result[:len(result) - self.backupCount]
        return result

    def emit(self, msg):
        """
        发送一个日志记录
        覆盖FileHandler中的emit方法，logging会自动调用此方法
        """
        try:
            if self.write_log_to_file():
                self.do_log_file_backup()
            logging.FileHandler.emit(self, msg)
        except (KeyboardInterrupt, SystemExit):
            raise
        except Exception as msg_ex:
            self.handleError(msg_ex)
